"""File 12panel_p.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 15 11:26:49 2020

 Example of a mixture of logit models, using Monte-Carlo integration.
 The datafile is organized as panel data.
 Three alternatives: Train, Car and Swissmetro
 SP data

The Swissmetro data is organized such that each row contains all the
responses of one individual.

"""
import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import (
    Beta,
    Variable,
    bioMultSum,
    bioDraws,
    MonteCarlo,
    log,
    exp,
)

from swissmetro import (
    database,
    NBR_QUESTIONS,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    TRAIN_AV_SP,
    SM_AV,
    CAR_AV_SP,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b12panel_p.py')

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 1, 0, None, 0)

# Define a random parameter, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# Definition of the utility functions
V1 = [
    ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED[q] + B_COST * TRAIN_COST_SCALED[q]
    for q in range(NBR_QUESTIONS)
]
V2 = [
    ASC_SM + B_TIME_RND * SM_TT_SCALED[q] + B_COST * SM_COST_SCALED[q]
    for q in range(NBR_QUESTIONS)
]
V3 = [
    ASC_CAR + B_TIME_RND * CAR_TT_SCALED[q] + B_COST * CAR_CO_SCALED[q]
    for q in range(NBR_QUESTIONS)
]

# Associate utility functions with the numbering of alternatives
V = [{1: V1[q], 2: V2[q], 3: V3[q]} for q in range(NBR_QUESTIONS)]

av = [{1: TRAIN_AV_SP[q], 2: SM_AV[q], 3: CAR_AV_SP[q]} for q in range(NBR_QUESTIONS)]


# Conditional to B_TIME_RND, the likelihood of one observation is
# given by the logit model (called the kernel)
obslogprob = [
    models.loglogit(V[q], av[q], Variable(f'CHOICE_{q}')) for q in range(NBR_QUESTIONS)
]

# Conditional to B_TIME_RND, the likelihood of all observations for
# one individual (the trajectory) is the product of the likelihood of
# each observation.
condprobIndiv = exp(bioMultSum(obslogprob))

# We integrate over B_TIME_RND using Monte-Carlo
logprob = log(MonteCarlo(condprobIndiv))

# Create the Biogeme object
biogeme = bio.BIOGEME(database, logprob)
biogeme.modelName = '12panel_p'

# Estimate the parameters.
results = biogeme.estimate()
pandasResults = results.getEstimatedParameters()
print(pandasResults)



""" File b01logit_all_algos.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:18:01 2023

 Logit model. Estimation with several algorithms.

"""
import biogeme.biogeme as bio
from biogeme import models
import biogeme.messaging as msg
from biogeme.expressions import Beta
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# Define level of verbosity
logger = msg.bioMessage()
logger.setSilent()
# logger.setWarning()
# logger.setGeneral()
# logger.setDetailed()

algos = {
    'scipy                ': 'scipy',
    'Line search          ': 'LS-newton',
    'Trust region (dogleg)': 'TR-newton',
    'Trust region (cg)    ': 'TR-newton',
    'LS-BFGS              ': 'LS-BFGS',
    'TR-BFGS              ': 'TR-BFGS',
    'Simple bounds Newton ': 'simple_bounds',
    'Simple bounds BFGS   ': 'simple_bounds',
    'Simple bounds hybrid ': 'simple_bounds',
}

algoParameters = {
    'Trust region (dogleg)': {'dogleg': True},
    'Trust region (cg)': {'dogleg': False},
    'Simple bounds Newton ': {'second_derivatives': 1.0},
    'Simple bounds BFGS   ': {'second_derivatives': 0.0},
    'Simple bounds hybrid ': {'second_derivatives': 0.5},
}

results = {}
print("Algorithm\t\tloglike\t\tnormg\ttime\t\tdiagnostic")
print("+++++++++\t\t+++++++\t\t+++++\t++++\t\t++++++++++")

for name, algo in algos.items():
    # Create the Biogeme object
    the_biogeme = bio.BIOGEME(database, logprob)
    the_biogeme.algorithm_name = algo
    the_biogeme.modelName = f'b01logit_all_algos_{name}'.strip()
    p = algoParameters.get(name)
    if p is not None:
        for attr, value in p.items():
            setattr(the_biogeme, attr, value)
    results[name] = the_biogeme.estimate()
    print(
        f'{name}\t{results[name].data.logLike:.2f}\t'
        f'{results[name].data.gradientNorm:.2g}\t'
        f'{results[name].data.optimizationMessages["Optimization time"]}'
        f'\t{results[name].data.optimizationMessages["Cause of termination"]}'
    )



"""File optima_results.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 16:40:23 2023

Assisted specification for the Optima case study. Generation of the
results from the Pareto file

"""
try: 
    import matplotlib.pyplot as plt
    can_plot = True
except ModuleNotFoundError:
    can_plot = False
import biogeme.logging as blog
from biogeme.assisted import AssistedSpecification, ParetoPostProcessing
from biogeme.results import compile_estimation_results
from biogeme.multiobjectives import loglikelihood_dimension
from optima_spec import the_biogeme

PARETO_FILE_NAME = 'optima.pareto'

screen_logger = blog.get_screen_logger(blog.INFO)
screen_logger.info('optima example: assisted specification')

nbr = the_biogeme.loglike.number_of_multiple_expressions()
if nbr is None:
    print('Number of possible specifications is too large to be calculated')

screen_logger.info('Prepare assisted specification')

assisted_specification = AssistedSpecification(
    biogeme_object=the_biogeme,
    multi_objectives=loglikelihood_dimension,
    pareto_file_name=PARETO_FILE_NAME,
)

post_processing = ParetoPostProcessing(
    biogeme_object=the_biogeme,
    pareto_file_name=PARETO_FILE_NAME,
)
post_processing.log_statistics()

all_results = post_processing.reestimate(recycle=True)

summary, description = compile_estimation_results(all_results, use_short_names=True)

print(summary)
for k, v in description.items():
    if k != v:
        print(f'{k}: {v} AIC={summary.at["Akaike Information Criterion", k]}')

if can_plot:
    ax = post_processing.plot()
    plt.show()



"""File b19individual_level_parameters

:author: Michel Bierlaire, EPFL
:date: Mon Apr 10 12:17:12 2023

Calculation of the individual level parameters for model 05normalMixture
"""

import os
import pickle
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, bioDraws, MonteCarlo

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# Values of the parameters estimated by the model 05normalMixture
ASC_CAR = Beta('ASC_CAR', 0.137, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', -0.402, None, None, 0)
B_COST = Beta('B_COST', -1.28, None, None, 0)

# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
B_TIME = Beta('B_TIME', -2.26, None, None, 0)

# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1.65, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

beta_values = {
    'ASC_CAR': 0.137,
    'ASC_TRAIN': -0.402,
    'B_COST': -1.28,
    'B_TIME': -2.26,
    'B_TIME_S': 1.65,
}

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Conditional to B_TIME_RND, we have a logit model (called the kernel)
prob_chosen = models.logit(V, av, CHOICE)

numerator = MonteCarlo(B_TIME_RND * prob_chosen)
denominator = MonteCarlo(prob_chosen)

simulate = {
    'Numerator': numerator,
    'Denominator': denominator,
    'Choice': CHOICE,
}

PICKLE_FILE = 'b19individual_level_parameters.pickle'
if os.path.isfile(PICKLE_FILE):
    with open(PICKLE_FILE, 'rb') as f:
        sim = pickle.load(f)
else:
    biosim = bio.BIOGEME(database, simulate)
    sim = biosim.simulate(beta_values)
    sim['Individual-level parameters'] = sim['Numerator'] / sim['Denominator']
    with open(PICKLE_FILE, 'wb') as f:
        pickle.dump(sim, f)

HTML_FILE = 'b19individual_level_parameters.html'
with open(HTML_FILE, 'w', encoding='utf-8') as h:
    print(sim.to_html(), file=h)
print(f'Simulated values available in {HTML_FILE}')



"""
Test the expressions module

:author: Michel Bierlaire
:data: Wed Apr 29 17:47:53 2020

"""
# Bug in pylint
# pylint: disable=no-member
#
# Too constraining
# pylint: disable=invalid-name, too-many-instance-attributes
#
# Not needed in test
# pylint: disable=missing-function-docstring, missing-class-docstring

import unittest
import numpy as np
import pandas as pd
import biogeme.exceptions as excep
import biogeme.expressions as ex
from biogeme import models
from biogeme.idmanager import IdManager
from biogeme.elementary_expressions import TypeOfElementaryExpression
from test_data import getData


class test_expressions(unittest.TestCase):
    def setUp(self):
        self.myData = getData(2)
        self.Person = ex.Variable('Person')
        self.Variable1 = ex.Variable('Variable1')
        self.Variable2 = ex.Variable('Variable2')
        self.Choice = ex.Variable('Choice')
        self.Av1 = ex.Variable('Av1')
        self.Av2 = ex.Variable('Av2')
        self.Av3 = ex.Variable('Av3')
        self.beta1 = ex.Beta('beta1', 0.2, None, None, 0)
        self.beta2 = ex.Beta('beta2', 0.4, None, None, 0)
        self.beta3 = ex.Beta('beta3', 1, None, None, 1)
        self.beta4 = ex.Beta('beta4', 0, None, None, 1)
        self.omega1 = ex.RandomVariable('omega1')
        self.omega2 = ex.RandomVariable('omega2')
        self.xi1 = ex.bioDraws('xi1', 'NORMAL')
        self.xi2 = ex.bioDraws('xi2', 'UNIF')
        self.xi3 = ex.bioDraws('xi3', 'WRONGTYPE')
        self.addTypeEqualityFunc(pd.DataFrame, self.assertDataframeEqual)

    def test_repr(self):
        result = repr(self.beta1)
        expected_result = 'beta1(init=0.2)'
        self.assertEqual(result, expected_result)

    def test_errors(self):
        with self.assertRaises(excep.BiogemeError):
            _ = ex.Numeric(1) / 'ert'

    #        with self.assertRaises(excep.BiogemeError):
    #            _ = 'ert' / Numeric(1)

    def assertDataframeEqual(self, a, b, msg):
        try:
            pd.testing.assert_frame_equal(a, b)
        except AssertionError as e:
            raise self.failureException(msg) from e

    def test_isNumeric(self):
        result = ex.isNumeric(1)
        self.assertTrue(result)
        result = ex.isNumeric(0.1)
        self.assertTrue(result)
        result = ex.isNumeric(True)
        self.assertTrue(result)
        result = ex.isNumeric(self)
        self.assertFalse(result)

    def test_add(self):
        result = self.Variable1 + self.Variable2
        self.assertEqual(str(result), '(Variable1 + Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 + 1
        self.assertEqual(str(result), '(Variable1 + `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 + self.Variable1
        self.assertEqual(str(result), '(`1.0` + Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self + self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 + self

    def test_sub(self):
        result = self.Variable1 - self.Variable2
        self.assertEqual(str(result), '(Variable1 - Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 - 1
        self.assertEqual(str(result), '(Variable1 - `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 - self.Variable1
        self.assertEqual(str(result), '(`1.0` - Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self - self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 - self

    def test_mul(self):
        result = self.Variable1 * self.Variable2
        self.assertEqual(str(result), '(Variable1 * Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 * 1
        self.assertEqual(str(result), '(Variable1 * `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 * self.Variable1
        self.assertEqual(str(result), '(`1.0` * Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self * self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 * self

    def test_div(self):
        result = self.Variable1 / self.Variable2
        self.assertEqual(str(result), '(Variable1 / Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 / 1
        self.assertEqual(str(result), '(Variable1 / `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 / self.Variable1
        self.assertEqual(str(result), '(`1.0` / Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self / self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 / self

    def test_neg(self):
        result = -self.Variable1
        self.assertEqual(str(result), '(-Variable1)')
        self.assertTrue(result.children[0] is self.Variable1)

    def test_pow(self):
        result = self.Variable1**self.Variable2
        self.assertEqual(str(result), '(Variable1 ** Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1**1
        self.assertEqual(str(result), '(Variable1 ** `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1**self.Variable1
        self.assertEqual(str(result), '(`1.0` ** Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self**self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1**self

    def test_and(self):
        result = self.Variable1 & self.Variable2
        self.assertEqual(str(result), '(Variable1 and Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 & 1
        self.assertEqual(str(result), '(Variable1 and `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 & self.Variable1
        self.assertEqual(str(result), '(`1.0` and Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self & self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 & self

    def test_or(self):
        result = self.Variable1 | self.Variable2
        self.assertEqual(str(result), '(Variable1 or Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 | 1
        self.assertEqual(str(result), '(Variable1 or `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 | self.Variable1
        self.assertEqual(str(result), '(`1.0` or Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self | self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 | self

    def test_eq(self):
        result = self.Variable1 == self.Variable2
        self.assertEqual(str(result), '(Variable1 == Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 == 1
        self.assertEqual(str(result), '(Variable1 == `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 == self.Variable1
        self.assertEqual(str(result), '(Variable1 == `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self == self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 == self

    def test_neq(self):
        result = self.Variable1 != self.Variable2
        self.assertEqual(str(result), '(Variable1 != Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 != 1
        self.assertEqual(str(result), '(Variable1 != `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 != self.Variable1
        self.assertEqual(str(result), '(Variable1 != `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self != self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 != self

    def test_le(self):
        result = self.Variable1 <= self.Variable2
        self.assertEqual(str(result), '(Variable1 <= Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 <= 1
        self.assertEqual(str(result), '(Variable1 <= `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 <= self.Variable1
        self.assertEqual(str(result), '(Variable1 >= `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self <= self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 <= self

    def test_ge(self):
        result = self.Variable1 >= self.Variable2
        self.assertEqual(str(result), '(Variable1 >= Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 >= 1
        self.assertEqual(str(result), '(Variable1 >= `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 >= self.Variable1
        self.assertEqual(str(result), '(Variable1 <= `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self >= self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 >= self

    def test_lt(self):
        result = self.Variable1 < self.Variable2
        self.assertEqual(str(result), '(Variable1 < Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 < 1
        self.assertEqual(str(result), '(Variable1 < `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 < self.Variable1
        self.assertEqual(str(result), '(Variable1 > `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self < self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 < self

    def test_gt(self):
        result = self.Variable1 > self.Variable2
        self.assertEqual(str(result), '(Variable1 > Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 > 1
        self.assertEqual(str(result), '(Variable1 > `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 > self.Variable1
        self.assertEqual(str(result), '(Variable1 < `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(excep.BiogemeError):
            result = self > self.Variable1

        with self.assertRaises(excep.BiogemeError):
            result = self.Variable1 > self

    def test_getValue_c(self):
        result = self.Variable1.getValue_c(database=self.myData, prepareIds=True)
        np.testing.assert_equal(result, [10, 20, 30, 40, 50])

    def test_DefineVariable(self):
        _ = self.myData.DefineVariable('newvar_b', self.Variable1 + self.Variable2)
        cols = self.myData.data.columns
        added = 'newvar_b' in cols
        self.assertTrue(added)
        self.myData.data['newvar_p'] = (
            self.myData.data['Variable1'] + self.myData.data['Variable2']
        )
        pd.testing.assert_series_equal(
            self.myData.data['newvar_b'],
            self.myData.data['newvar_p'],
            check_dtype=False,
            check_names=False,
        )

    def test_expr1(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        self.assertAlmostEqual(expr1.getValue(), -1.275800115089098, 5)
        res = expr1.getValue_c(prepareIds=True)
        self.assertAlmostEqual(res, -1.275800115089098, 5)

    def test_expr1_newvalues(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        newvalues = {'beta1': 1, 'beta2': 2, 'beta3': 3, 'beta4': 2}
        expr1.change_init_values(newvalues)
        self.assertAlmostEqual(expr1.getValue(), 1.9323323583816936, 5)
        res = expr1.getValue_c(prepareIds=True)
        self.assertAlmostEqual(res, 1.9323323583816936, 5)

    def test_expr1_derivatives(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        newvalues = {'beta1': 1, 'beta2': 2, 'beta3': 3, 'beta4': 2}
        expr1.change_init_values(newvalues)
        f, g, h, b = expr1.getValueAndDerivatives(prepareIds=True)
        self.assertAlmostEqual(f, 1.9323323583816936, 5)
        g_ok = [2.0, 0.10150146242745953]
        h0_ok = [0.0, 0.0]
        h1_ok = [0.0, -0.16916910404576588]
        bhhh0_ok = [4, 0.20300292]
        bhhh1_ok = [0.20300292, 0.01030255]
        for check_left, check_right in zip(g, g_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(h[0], h0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(h[1], h1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(b[0], bhhh0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(b[1], bhhh1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

    def test_expr1_gradient(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        f, g, h, b = expr1.getValueAndDerivatives(
            gradient=True, hessian=False, bhhh=False, prepareIds=True
        )
        self.assertAlmostEqual(f, -1.275800115089098, 5)
        g_ok = [2.0, 5.865300402811844]
        for check_left, check_right in zip(g, g_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        self.assertIsNone(h)
        self.assertIsNone(b)

    def test_expr1_function(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        the_function = expr1.createFunction()
        f, g, h = the_function([1, 2])
        self.assertAlmostEqual(f, 1.9323323583816936, 5)
        g_ok = [2.0, 0.10150146242745953]
        h0_ok = [0.0, 0.0]
        h1_ok = [0.0, -0.16916910404576588]
        for check_left, check_right in zip(g, g_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(h[0], h0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(h[1], h1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

        f, g, h = the_function([10, -2])
        self.assertAlmostEqual(f, 23.694528049465326, 5)
        g_ok = [2, -1.84726402]
        h0_ok = [0.0, 0.0]
        h1_ok = [0.0, 1.84726402]
        for check_left, check_right in zip(g, g_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(h[0], h0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(h[1], h1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

    def test_expr1_database(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        f_list, g_list, h_list, bhhh_list = expr1.getValueAndDerivatives(
            database=self.myData, aggregation=False, prepareIds=True
        )
        f_ok = -1.27580012
        g_ok = [2.0, 5.8653004]
        h0_ok = [0.0, 0.0]
        h1_ok = [0.0, -31.00230213]
        bhhh0_ok = [4.0, 11.73060081]
        bhhh1_ok = [11.73060081, 34.40174882]
        for i, f in enumerate(f_list):
            self.assertAlmostEqual(f, f_ok, 5)
            for check_left, check_right in zip(g_list[i], g_ok):
                self.assertAlmostEqual(check_left, check_right, 5)
            for check_left, check_right in zip(h_list[i][0], h0_ok):
                self.assertAlmostEqual(check_left, check_right, 5)
            for check_left, check_right in zip(h_list[i][1], h1_ok):
                self.assertAlmostEqual(check_left, check_right, 5)
            for check_left, check_right in zip(bhhh_list[i][0], bhhh0_ok):
                self.assertAlmostEqual(check_left, check_right, 5)
            for check_left, check_right in zip(bhhh_list[i][1], bhhh1_ok):
                self.assertAlmostEqual(check_left, check_right, 5)

    def test_expr1_database_agg(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        f, g, h, b = expr1.getValueAndDerivatives(
            database=self.myData, aggregation=True, prepareIds=True
        )
        f_ok = -6.37900057544549
        g_ok = [10.0, 29.32650201405922]
        h0_ok = [0.0, 0.0]
        h1_ok = [0.0, -155.01151064574157]
        b0_ok = [20.0, 58.65300403]
        b1_ok = [58.65300403, 172.00874408]
        self.assertAlmostEqual(f, f_ok, 5)
        for check_left, check_right in zip(g, g_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(h[0], h0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(h[1], h1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(b[0], b0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(b[1], b1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

    def test_setOfBetas(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        s = expr1.set_of_elementary_expression(
            the_type=TypeOfElementaryExpression.FREE_BETA
        )
        self.assertSetEqual(s, {'beta1', 'beta2'})
        s = expr1.set_of_elementary_expression(
            the_type=TypeOfElementaryExpression.FIXED_BETA
        )
        self.assertSetEqual(s, {'beta3'})
        s = expr1.set_of_elementary_expression(the_type=TypeOfElementaryExpression.BETA)
        self.assertSetEqual(s, {'beta1', 'beta2', 'beta3'})

    def test_setOfVariables(self):
        expr1 = 2 * self.Variable1 - ex.exp(-self.Variable2) / (
            self.Variable1 * (self.Variable1 >= self.Variable2)
        )
        s = expr1.set_of_elementary_expression(
            the_type=TypeOfElementaryExpression.VARIABLE
        )
        self.assertSetEqual(s, {'Variable1', 'Variable2'})

    def test_getElementaryExpression(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        ell = expr1.getElementaryExpression('beta2')
        self.assertEqual(ell.name, 'beta2')

    def test_expr2(self):
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        with self.assertRaises(excep.BiogemeError):
            expr2.getValue()
        with self.assertRaises(excep.BiogemeError):
            expr2.getValue_c()
        res = list(expr2.getValue_c(database=self.myData, prepareIds=True))
        self.assertListEqual(res, [4.0, 8.0, 12.0, 16.0, 20.0])

    def test_dictOfBetas(self):
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (self.beta3 * (self.beta2 >= self.beta1))
        b = expr2.dict_of_elementary_expression(
            the_type=TypeOfElementaryExpression.BETA
        )
        # Note that the following checks only the labels. Its probably
        # good enough for our purpose.
        self.assertDictEqual(b, {'beta1': 0, 'beta2': self.beta2, 'beta3': self.beta3})

    def test_dictOfVariables(self):
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (self.beta3 * (self.beta2 >= self.beta1))
        b = expr2.dict_of_elementary_expression(
            the_type=TypeOfElementaryExpression.VARIABLE
        )
        # Note that the following checks only the labels. Its probably
        # good enough for our purpose.
        self.assertDictEqual(
            b, {'Variable1': self.Variable1, 'Variable2': self.Variable2}
        )

    def test_dictOfRandomVariables(self):
        expr = -(self.omega1 + self.omega2 + self.Variable1)
        b = expr.dict_of_elementary_expression(
            the_type=TypeOfElementaryExpression.RANDOM_VARIABLE
        )
        self.assertDictEqual(b, {'omega1': self.omega1, 'omega2': self.omega2})

    def test_dictOfDraws(self):
        expr = -(self.xi1 + self.xi2 - self.xi3 + self.Variable1)
        b = expr.dict_of_elementary_expression(
            the_type=TypeOfElementaryExpression.DRAWS
        )
        self.assertDictEqual(b, {'xi1': 'NORMAL', 'xi2': 'UNIF', 'xi3': 'WRONGTYPE'})

    def test_getClassName(self):
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (self.beta3 * (self.beta2 >= self.beta1))
        c = expr2.getClassName()
        self.assertEqual(c, 'Minus')

    def test_getSignature(self):
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (self.beta3 * (self.beta2 >= self.beta1))
        expr2.setIdManager(IdManager([expr2], self.myData, 0))
        s = expr2.getSignature()
        self.assertEqual(len(s), 17)

    def test_embedExpression(self):
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (self.beta3 * (self.beta2 >= self.beta1))
        self.assertTrue(expr2.embedExpression('Minus'))
        self.assertFalse(expr2.embedExpression('bioDraws'))

    def test_panel_variables(self):
        expr_ok = ex.PanelLikelihoodTrajectory(self.Variable1)
        check_ok = expr_ok.check_panel_trajectory()
        self.assertEqual(len(check_ok), 0)
        expr_not_ok = self.Variable2 + ex.PanelLikelihoodTrajectory(self.Variable1)
        check_not_ok = expr_not_ok.check_panel_trajectory()
        self.assertSetEqual(check_not_ok, {'Variable2'})

    def test_check_draws(self):
        d1 = ex.bioDraws('d1', 'UNIFORM')
        d2 = ex.bioDraws('d2', 'UNIFORM')
        expr_ok = ex.MonteCarlo(d1 * d2)
        check_ok = expr_ok.check_draws()
        self.assertEqual(len(check_ok), 0)
        expr_not_ok = d1 + ex.MonteCarlo(d1 * d2)
        check_not_ok = expr_not_ok.check_draws()
        self.assertSetEqual(check_not_ok, {'d1'})

    def test_check_rv(self):
        d1 = ex.RandomVariable('d1')
        d2 = ex.RandomVariable('d2')
        expr_ok = ex.Integrate(1, d1)
        check_ok = expr_ok.check_rv()
        self.assertEqual(len(check_ok), 0)
        expr_not_ok = d1 + ex.Integrate(1, d2)
        check_not_ok = expr_not_ok.check_rv()
        self.assertSetEqual(check_not_ok, {'d1'})

    def test_countPanelTrajectoryExpressions(self):
        expr1 = self.beta1
        c1 = expr1.countPanelTrajectoryExpressions()
        self.assertEqual(c1, 0)
        expr2 = ex.PanelLikelihoodTrajectory(self.beta1)
        c2 = expr2.countPanelTrajectoryExpressions()
        self.assertEqual(c2, 1)
        expr3 = ex.PanelLikelihoodTrajectory(ex.PanelLikelihoodTrajectory(self.beta1))
        c3 = expr3.countPanelTrajectoryExpressions()
        expr4 = self.Variable1 + ex.PanelLikelihoodTrajectory(
            ex.PanelLikelihoodTrajectory(self.beta1)
        )
        c4 = expr4.countPanelTrajectoryExpressions()
        self.assertEqual(c4, 2)

    def test_ids_multiple_formulas(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        collectionOfFormulas = [expr1, expr2]
        formulas = IdManager(collectionOfFormulas, self.myData, None)
        elementaryExpressionIndex_ok = {
            'beta1': 0,
            'beta2': 1,
            'beta3': 2,
            'beta4': 3,
            'Person': 4,
            'Exclude': 5,
            'Variable1': 6,
            'Variable2': 7,
            'Choice': 8,
            'Av1': 9,
            'Av2': 10,
            'Av3': 11,
        }
        self.assertDictEqual(
            formulas.elementary_expressions.indices,
            elementaryExpressionIndex_ok,
        )
        self.assertListEqual(
            list(formulas.free_betas.expressions.keys()), ['beta1', 'beta2']
        )
        self.assertListEqual(formulas.free_betas.names, ['beta1', 'beta2'])
        self.assertListEqual(
            list(formulas.fixed_betas.expressions.keys()), ['beta3', 'beta4']
        )
        self.assertListEqual(formulas.fixed_betas.names, ['beta3', 'beta4'])
        self.assertFalse(formulas.random_variables.expressions)
        self.assertFalse(formulas.draws.expressions)

    def test_expr3(self):
        myDraws = ex.bioDraws('myDraws', 'UNIFORM')
        expr3 = ex.MonteCarlo(myDraws * myDraws)
        with self.assertRaises(excep.BiogemeError):
            res = expr3.getValue_c(numberOfDraws=100000)
        res = expr3.getValue_c(
            database=self.myData, numberOfDraws=100000, prepareIds=True
        )
        for v in res:
            self.assertAlmostEqual(v, 1.0 / 3.0, 2)

    def test_expr4(self):
        omega = ex.RandomVariable('omega')
        a = 0
        b = 1
        x = a + (b - a) / (1 + ex.exp(-omega))
        dx = (b - a) * ex.exp(-omega) * (1 + ex.exp(-omega)) ** (-2)
        integrand = x * x
        expr4 = ex.Integrate(integrand * dx / (b - a), 'omega')
        res = expr4.getValue_c(database=self.myData, prepareIds=True)
        for v in res:
            self.assertAlmostEqual(v, 1.0 / 3.0, 2)

    def test_expr5(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (self.beta3 * (self.beta2 >= self.beta1))
        expr5 = ex.Elem({1: expr1, 2: expr2}, self.Person) / 10
        res = list(expr5.getValue_c(database=self.myData, prepareIds=True))
        res_ok = [
            -0.02703200460356393,
            -0.02703200460356393,
            -0.02703200460356393,
            1.6,
            2.0,
        ]
        for check_left, check_right in zip(res, res_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

    def test_expr6(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (self.beta3 * (self.beta2 >= self.beta1))
        omega = ex.RandomVariable('omega')
        a = 0
        b = 1
        x = a + (b - a) / (1 + ex.exp(-omega))
        dx = (b - a) * ex.exp(-omega) * (1 + ex.exp(-omega)) ** (-2)
        integrand = x * x
        expr4 = ex.Integrate(integrand * dx / (b - a), 'omega')
        expr6 = ex.bioMultSum([expr1, expr2, expr4])
        res = list(expr6.getValue_c(database=self.myData, prepareIds=True))
        res_ok = [
            4.063012266030643,
            8.063012266030643,
            12.063012266030643,
            16.063012266030643,
            20.063012266030643,
        ]
        for check_left, check_right in zip(res, res_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

    def test_expr7(self):
        V = {0: -self.beta1, 1: -self.beta2, 2: -self.beta1}
        av = {0: 1, 1: 1, 2: 1}
        expr7 = ex.LogLogit(V, av, 1)
        r = expr7.getValue()
        self.assertAlmostEqual(r, -1.2362866960692134, 5)
        expr8 = models.loglogit(V, av, 1)
        res = expr8.getValue_c(database=self.myData, prepareIds=True)
        for v in res:
            self.assertAlmostEqual(v, -1.2362866960692136, 5)

    def test_expr9(self):
        V = {0: -self.beta1, 1: -self.beta2, 2: -self.beta1}
        av = {0: 1, 1: 1, 2: 1}
        expr8 = models.loglogit(V, av, 1)
        expr9 = ex.Derive(expr8, 'beta2')
        res = expr9.getValue_c(database=self.myData, prepareIds=True)
        for v in res:
            self.assertAlmostEqual(v, -0.7095392129298093, 5)

    def test_expr10(self):
        expr10 = ex.bioNormalCdf(self.Variable1 / 10 - 1)
        res = expr10.getValue_c(database=self.myData, prepareIds=True)
        for i, j in zip(
            res,
            [
                0.5,
                0.8413447460685283,
                0.9772498680518218,
                0.99865010196837,
                0.9999683287581669,
            ],
        ):
            self.assertAlmostEqual(i, j, 5)

    def test_expr11(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (self.beta3 * (self.beta2 >= self.beta1))
        expr5 = ex.Elem({1: expr1, 2: expr2}, self.Person) / 10
        expr10 = ex.bioNormalCdf(self.Variable1 / 10 - 1)
        expr11 = ex.bioMin(expr5, expr10)
        res = expr11.getValue_c(database=self.myData, prepareIds=True)
        res_ok = [-0.027032, -0.027032, -0.027032, 0.9986501, 0.99996833]
        for i, j in zip(
            res,
            res_ok,
        ):
            self.assertAlmostEqual(i, j, 5)

    def test_expr12(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(
            -self.beta2 * self.Variable2
        ) / (self.beta3 * (self.beta2 >= self.beta1))
        expr5 = ex.Elem({1: expr1, 2: expr2}, self.Person) / 10
        expr10 = ex.bioNormalCdf(self.Variable1 / 10 - 1)
        expr12 = ex.bioMax(expr5, expr10)
        res = expr12.getValue_c(database=self.myData, prepareIds=True)
        for i, j in zip(res, [0.5, 0.8413447460685283, 0.9772498680518218, 1.6, 2.0]):
            self.assertAlmostEqual(i, j, 5)

    def test_expr13(self):
        newvar = self.myData.DefineVariable('newvar', self.Variable1 + self.Variable2)
        terms = [
            (self.beta1, ex.Variable('Variable1')),
            (self.beta2, ex.Variable('Variable2')),
            (self.beta3, newvar),
        ]
        expr13 = ex.bioLinearUtility(terms)
        res = expr13.getValue_c(database=self.myData, prepareIds=True)
        res_ok = [
            152,
            304,
            456,
            608,
            760,
        ]
        for i, j in zip(res, res_ok):
            self.assertAlmostEqual(i, j, 5)
        expr13bis = (
            self.beta1 * ex.Variable('Variable1')
            + self.beta2 * ex.Variable('Variable2')
            + self.beta3 * newvar
        )
        res = expr13bis.getValue_c(database=self.myData, prepareIds=True)
        res_ok = [
            152,
            304,
            456,
            608,
            760,
        ]
        for i, j in zip(res, res_ok):
            self.assertAlmostEqual(i, j, 5)

    def test_panel(self):
        ss = self.myData.getSampleSize()
        self.assertEqual(ss, 5)
        self.myData.panel('Person')
        ss = self.myData.getSampleSize()
        self.assertEqual(ss, 2)

    def test_expr14(self):
        c1 = ex.bioDraws('draws1', 'NORMAL_HALTON2')
        c2 = ex.bioDraws('draws2', 'NORMAL_HALTON2')
        U1 = ex.Beta('beta1', 0, None, None, 0) * ex.Variable('Variable1') + 10 * c1
        U2 = ex.Beta('beta2', 0, None, None, 0) * ex.Variable('Variable2') + 10 * c2
        U3 = 0
        U = {1: U1, 2: U2, 3: U3}
        av = {1: self.Av1, 2: self.Av2, 3: self.Av3}
        expr14 = ex.log(
            ex.MonteCarlo(
                ex.PanelLikelihoodTrajectory(models.logit(U, av, self.Choice))
            )
        )
        self.myData.panel('Person')
        res = expr14.getValue_c(
            database=self.myData, numberOfDraws=100000, prepareIds=True
        )
        res_ok = [-3.93304323, -2.10368902]
        for i, j in zip(res, res_ok):
            self.assertAlmostEqual(i, j, 3)

        f_list, g_list, h_list, b_list = expr14.getValueAndDerivatives(
            database=self.myData,
            numberOfDraws=10000000,
            gradient=True,
            hessian=True,
            bhhh=True,
            aggregation=False,
            prepareIds=True,
        )
        g_flat = g_list.flatten() / 100
        h_flat = h_list.flatten() / 1000
        b_flat = b_list.flatten() / 1000
        f_ok = [-3.93304323, -2.10368902]
        for i, j in zip(f_list, f_ok):
            self.assertAlmostEqual(i, j, 3)
        g_ok = [-0.12583932, 0.74160679, -0.03182045, 0.68179551]
        for i, j in zip(g_flat, g_ok):
            self.assertAlmostEqual(i, j, 2)
        h_ok = [
            -0.16720814,
            1.59974094,
            1.59974094,
            -16.72081405,
            -0.98711321,
            9.80068839,
            9.80068839,
            -98.71132074,
        ]
        for i, j in zip(h_flat, h_ok):
            self.assertAlmostEqual(i, j, 2)
        b_ok = [
            0.15835535,
            -0.93323295,
            -0.93323295,
            5.49980634,
            0.01012541,
            -0.21695039,
            -0.21695039,
            4.64845114,
        ]
        for i, j in zip(b_flat, b_ok):
            self.assertAlmostEqual(i, j, 2)


if __name__ == '__main__':
    unittest.main()



"""File b04validation.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:24:32 2023

 Example of the out-of-sample validation of a logit model.
"""
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b04validation'

# Estimate the parameters
results = the_biogeme.estimate()

# The validation consists in organizing the data into several slices
# of about the same size, randomly defined. Each slice is considered
# as a validation dataset. The model is then re-estimated using all
# the data except the slice, and the estimated model is applied on the
# validation set (i.e. the slice). The value of the log likelihood for
# each observation in the validation set is reported in a
# dataframe. As this is done for each slice, the output is a list of
# dataframes, each corresponding to one of these exercises.

validationData = database.split(slices=5)

validation_results = the_biogeme.validate(results, validationData)

for slide in validation_results:
    print(
        f'Log likelihood for {slide.shape[0]} validation data: '
        f'{slide["Loglikelihood"].sum()}'
    )



"""File b11cnl.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:06:44 2023

 Example of a cross-nested logit model.
 Three alternatives: Train, Car and Swissmetro
 Train and car are in the same nest.
"""

import biogeme.logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b11cnl.py')

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

MU_EXISTING = Beta('MU_EXISTING', 1, 1, 10, 0)
MU_PUBLIC = Beta('MU_PUBLIC', 1, 1, 10, 0)
ALPHA_EXISTING = Beta('ALPHA_EXISTING', 0.5, 0, 1, 0)
ALPHA_PUBLIC = 1 - ALPHA_EXISTING

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of nests
# Nest membership parameters
alpha_existing = {1: ALPHA_EXISTING, 2: 0.0, 3: 1.0}

alpha_public = {1: ALPHA_PUBLIC, 2: 1.0, 3: 0.0}

nest_existing = MU_EXISTING, alpha_existing
nest_public = MU_PUBLIC, alpha_public
nests = nest_existing, nest_public

# The choice model is a cross-nested logit, with availability conditions
logprob = models.logcnl_avail(V, av, nests, CHOICE)

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b11cnl'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""Processing of the choice data with sampling of alternatives

:author: Michel Bierlaire
:date: Sat Apr 15 16:46:42 2023

"""
import pandas as pd
import biogeme.database as db
from biogeme.expressions import Variable, log

from sample import (
    FILE_NAME,
    SAMPLE_SIZE as N_ALT,
)

df = pd.read_csv(FILE_NAME)
database = db.Database('choice_data', df)

# We need to calculate the distances between each individual and each
# restaurant in the sampled choice set.

for i in range(N_ALT):
    database.DefineVariable(
        f'log_dist_{i}',
        log(
            (
                (Variable('user_lat') - Variable(f'rest_lat_{i}')) ** 2
                + (Variable('user_lon') - Variable(f'rest_lon_{i}')) ** 2
            )
            ** 0.5
        ),
    )



"""Generation of the default parameter file and values

:author: Michel Bierlaire
:date: Wed Nov 30 10:14:35 2022

IMPORTANT: when only one "check" function is provided, do not forget
to insert a comma at the end, before the closing parenthesis for the
tuple.
See https://www.w3schools.com/python/gloss_python_tuple_one_item.asp
"""

from typing import NamedTuple, Union, Type, Callable
import numpy as np
import biogeme.optimization as opt
import biogeme.check_parameters as cp


class ParameterTuple(NamedTuple):
    name: str
    value: Union[bool, int, float, str]
    type: Type
    section: str
    description: str
    check: Callable


all_parameters_tuple = (
    ParameterTuple(
        name='identification_threshold',
        value=1.0e-5,
        type=float,
        section='Output',
        description=(
            'float: if the smallest eigenvalue of the second derivative '
            'matrix is lesser or equal to this parameter, the model is '
            'considered not identified. The corresponding eigenvector '
            'is then reported to identify the parameters involved in the issue.'
        ),
        check=(cp.is_number,),
    ),
    ParameterTuple(
        name='only_robust_stats',
        value=True,
        type=bool,
        section='Output',
        description=(
            'bool: "True" if only the robust statistics need to be reported.'
            ' If "False", the statistics from the Rao-Cramer bound are '
            'also reported.'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='generate_html',
        value=True,
        type=bool,
        section='Output',
        description=(
            'bool: "True" if the HTML file with the results must be generated.'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='generate_pickle',
        value=True,
        type=bool,
        section='Output',
        description=(
            'bool: "True" if the pickle file with the ' 'results must be generated.'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='number_of_threads',
        value=0,
        type=int,
        section='MultiThreading',
        description=(
            'int: Number of threads/processors to be used. If'
            ' the parameter is 0, the number of'
            ' available threads is calculated using'
            ' cpu_count().'
        ),
        check=(cp.is_integer, cp.is_non_negative),
    ),
    ParameterTuple(
        name='number_of_draws',
        value=20000,
        type=int,
        section='MonteCarlo',
        description=('int: Number of draws for Monte-Carlo integration.'),
        check=(cp.is_integer, cp.is_positive),
    ),
    ParameterTuple(
        name='missing_data',
        value=99999,
        type=int,
        section='Specification',
        description=(
            'number: If one variable has this value, it is assumed'
            ' that a data is missing and an exception will'
            ' be triggered.'
        ),
        check=(cp.is_number,),
    ),
    ParameterTuple(
        name='seed',
        value=0,
        type=int,
        section='MonteCarlo',
        description=(
            'int: Seed used for the pseudo-random number generation. It is'
            ' useful only when each run should generate the exact same'
            ' result. If 0, a new seed is used at each run.'
        ),
        check=(cp.is_integer, cp.is_non_negative),
    ),
    ParameterTuple(
        name='bootstrap_samples',
        value=100,
        type=int,
        section='Estimation',
        description=('int: number of re-estimations for bootstrap sampling.'),
        check=(cp.is_integer, cp.is_non_negative),
    ),
    ParameterTuple(
        name='max_number_parameters_to_report',
        value=15,
        type=int,
        section='Estimation',
        description=(
            'int: maximum number of parameters to report during the estimation.'
        ),
        check=(cp.is_integer, cp.is_non_negative),
    ),
    ParameterTuple(
        name='save_iterations',
        value=True,
        type=bool,
        section='Estimation',
        description=(
            'bool: If True, the current iterate is saved after'
            ' each iteration, in a file named'
            ' ``__[modelName].iter``, where'
            ' ``[modelName]`` is the name given to the'
            ' model. If such a file exists, the starting'
            ' values for the estimation are replaced by'
            ' the values saved in the file.'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='maximum_number_catalog_expressions',
        value=100,
        type=int,
        section='Estimation',
        description=(
            'If the expression contrains catalogs, the parameter sets an '
            'upper bound of the total number of possible combinations '
            'that can be estimated in the same loop.'
        ),
        check=(
            cp.is_integer,
            cp.is_positive,
        ),
    ),
    ParameterTuple(
        name='optimization_algorithm',
        value='simple_bounds',
        type=str,
        section='Estimation',
        description=(
            f'str: optimization algorithm to be used for estimation. '
            f'Valid values: {list(opt.algorithms.keys())}'
        ),
        check=(cp.check_algo_name,),
    ),
    ParameterTuple(
        name='second_derivatives',
        value=1.0,
        type=float,
        section='SimpleBounds',
        description=(
            'float: proportion (between 0 and 1) of iterations when the '
            'analytical Hessian is calculated'
        ),
        check=(cp.zero_one, cp.is_number),
    ),
    ParameterTuple(
        name='tolerance',
        value=np.finfo(np.float64).eps ** 0.3333,
        type=float,
        section='SimpleBounds',
        description='float: the algorithm stops when this precision is reached',
        check=(cp.is_number,),
    ),
    ParameterTuple(
        name='max_iterations',
        value=100,
        type=int,
        section='SimpleBounds',
        description='int: maximum number of iterations',
        check=(cp.is_integer, cp.is_positive),
    ),
    ParameterTuple(
        name='infeasible_cg',
        value=False,
        type=bool,
        section='SimpleBounds',
        description=(
            'If True, the conjugate gradient algorithm may generate '
            'infeasible solutiona until termination.  The result will '
            'then be projected on the feasible domain.  If False, the '
            'algorithm stops as soon as an infeasible iterate is '
            'generated'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='initial_radius',
        value=10,
        type=float,
        section='SimpleBounds',
        description='Initial radius of the trust region',
        check=(cp.is_number, cp.is_positive),
    ),
    ParameterTuple(
        name='steptol',
        value=1.0e-5,
        type=float,
        section='SimpleBounds',
        description=(
            'The algorithm stops when the relative change in x is below '
            'this threshold. Basically, if p significant digits of x '
            'are needed, steptol should be set to 1.0e-p.'
        ),
        check=(cp.is_number, cp.is_positive),
    ),
    ParameterTuple(
        name='enlarging_factor',
        value=10,
        type=float,
        section='SimpleBounds',
        description=(
            'If an iteration is very successful, the radius of '
            'the trust region is multiplied by this factor'
        ),
        check=(cp.is_number, cp.is_positive),
    ),
    ParameterTuple(
        name='dogleg',
        value=True,
        type=bool,
        section='TrustRegion',
        description=(
            'bool: choice of the method to solve the trust region subproblem. '
            'True: dogleg. False: truncated conjugate gradient.'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='number_of_neighbors',
        value=20,
        type=int,
        section='AssistedSpecification',
        description=(
            'int: maximum number of neighbors that are visited by the VNS algorithm.'
        ),
        check=(cp.is_integer, cp.is_non_negative),
    ),
    ParameterTuple(
        name='largest_neighborhood',
        value=20,
        type=int,
        section='AssistedSpecification',
        description=(
            'int: size of the largest neighborhood copnsidered by the Variable '
            'Neighborhood Search (VNS) algorithm.'
        ),
        check=(cp.is_integer, cp.is_non_negative),
    ),
    ParameterTuple(
        name='maximum_attempts',
        value=100,
        type=int,
        section='AssistedSpecification',
        description=(
            'int: an attempts consists in selecting a solution in the Pareto '
            'set, and trying to improve it. The parameter imposes an upper bound '
            'on the total number of attempts, irrespectively if they are '
            'successful or not.'
        ),
        check=(cp.is_integer, cp.is_non_negative),
    ),
)



"""File optima_data.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 11:26:56 2023

Data for the Optima case study
"""

import pandas as pd
import biogeme.database as db
from biogeme.expressions import Variable

# Read the data
df = pd.read_csv('optima.dat', sep='\t')

df.loc[df['OccupStat'] > 2, 'OccupStat'] = 3
df.loc[df['OccupStat'] == -1, 'OccupStat'] = 3

df.loc[df['Education'] <= 3, 'Education'] = 3
df.loc[df['Education'] <= 3, 'Education'] = 3
df.loc[df['Education'] == 5, 'Education'] = 4
df.loc[df['Education'] == 8, 'Education'] = 7

df.loc[df['TripPurpose'] != 1, 'TripPurpose'] = 2

df.loc[df['CarAvail'] != 3, 'CarAvail'] = 1

database = db.Database('optima', df)

Choice = Variable('Choice')
CostCarCHF = Variable('CostCarCHF')
CarAvail = Variable('CarAvail')
HalfFareST = Variable('HalfFareST')
LineRelST = Variable('LineRelST')
AreaRelST = Variable('AreaRelST')
OtherST = Variable('OtherST')
GenAbST = Variable('GenAbST')
TimePT = Variable('TimePT')
TimeCar = Variable('TimeCar')
MarginalCostPT = Variable('MarginalCostPT')
distance_km = Variable('distance_km')
WaitingTimePT = Variable('WaitingTimePT')
TripPurpose = Variable('TripPurpose')
UrbRur = Variable('UrbRur')
LangCode = Variable('LangCode')
Gender = Variable('Gender')
OccupStat = Variable('OccupStat')
subscription = Variable('subscription')
CarAvail = Variable('CarAvail')
Education = Variable('Education')
NbTransf = Variable('NbTransf')
age = Variable('age')

exclude = ((Choice == -1) + (CostCarCHF < 0) + (CarAvail == 3) * (Choice == 1)) > 0
database.remove(exclude)

# Definition of new variables

otherSubscription = database.DefineVariable(
    'otherSubscription',
    ((HalfFareST == 1) + (LineRelST == 1) + (AreaRelST == 1) + (OtherST) == 1) > 0,
)

subscription = database.DefineVariable(
    'subscription', (GenAbST == 1) * 1 + (GenAbST != 1) * otherSubscription * 2
)

TimePT_scaled = database.DefineVariable('TimePT_scaled', TimePT / 200)
TimeCar_scaled = database.DefineVariable('TimeCar_scaled', TimeCar / 200)
MarginalCostPT_scaled = MarginalCostPT / 10
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5



"""File airline.py

:author: Michel Bierlaire, EPFL
:date: Fri Mar 31 10:45:43 2023

Assisted specification for the airline cases tudy
"""


import biogeme.logging as blog
from biogeme.assisted import AssistedSpecification
from biogeme.results import compile_estimation_results
from biogeme.multiobjectives import loglikelihood_dimension

from airline_spec import the_biogeme

screen_logger = blog.get_screen_logger(blog.INFO)
screen_logger.info('airline example: assisted specification')

PARETO_FILE_NAME = 'airline.pareto'

nbr = the_biogeme.loglike.number_of_multiple_expressions()
print('The number of possible specifications is {nbr}')

screen_logger.info('Prepare assisted specification')

assisted_specification = AssistedSpecification(
    biogeme_object=the_biogeme,
    multi_objectives=loglikelihood_dimension,
    pareto_file_name=PARETO_FILE_NAME,
)

screen_logger.info('Run assisted specification')

non_dominated_models = assisted_specification.run()

summary, description = compile_estimation_results(
    non_dominated_models, use_short_names=True
)
print(summary)
for k, v in description.items():
    if k != v:
        print(f'{k}: {v}')



import os
import unittest
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import (
    Beta,
    log,
    bioDraws,
    MonteCarlo,
    PanelLikelihoodTrajectory,
)


myPath = os.path.dirname(os.path.abspath(__file__))
df = pd.read_csv(f'{myPath}/swissmetro.dat', sep='\t')
database = db.Database('swissmetro', df)
database.panel('ID')

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
# print(database.data.describe())

globals().update(database.variables)

exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
database.remove(exclude)


ASC_CAR = Beta('ASC_CAR', 0.136, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', -1, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', -6.3, None, 0, 0)
B_COST = Beta('B_COST', -3.29, None, 0, 0)

SIGMA_CAR = Beta('SIGMA_CAR', 3.7, None, None, 0)
SIGMA_SM = Beta('SIGMA_SM', 0.759, None, None, 0)
SIGMA_TRAIN = Beta('SIGMA_TRAIN', 3.02, None, None, 0)

# Define a random parameter, normally distirbuted, designed to be used
# for Monte-Carlo simulation
EC_CAR = SIGMA_CAR * bioDraws('EC_CAR', 'NORMAL')
EC_SM = SIGMA_SM * bioDraws('EC_SM', 'NORMAL')
EC_TRAIN = SIGMA_TRAIN * bioDraws('EC_TRAIN', 'NORMAL')

SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

TRAIN_TT_SCALED = database.DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0)
TRAIN_COST_SCALED = database.DefineVariable(
    'TRAIN_COST_SCALED', TRAIN_COST / 100
)
SM_TT_SCALED = database.DefineVariable('SM_TT_SCALED', SM_TT / 100.0)
SM_COST_SCALED = database.DefineVariable('SM_COST_SCALED', SM_COST / 100)
CAR_TT_SCALED = database.DefineVariable('CAR_TT_SCALED', CAR_TT / 100)
CAR_CO_SCALED = database.DefineVariable('CAR_CO_SCALED', CAR_CO / 100)

# For latent class 1, whete the time coefficient is zero
V11 = ASC_TRAIN + B_COST * TRAIN_COST_SCALED + EC_TRAIN
V12 = ASC_SM + B_COST * SM_COST_SCALED + EC_SM
V13 = ASC_CAR + B_COST * CAR_CO_SCALED + EC_CAR

V1 = {1: V11, 2: V12, 3: V13}

# For latent class 2, whete the time coefficient is estimated
V21 = (
    ASC_TRAIN
    + B_TIME * TRAIN_TT_SCALED
    + B_COST * TRAIN_COST_SCALED
    + EC_TRAIN
)
V22 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED + EC_SM
V23 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED + EC_CAR

V2 = {1: V21, 2: V22, 3: V23}


# Associate the availability conditions with the alternatives

CAR_AV_SP = database.DefineVariable('CAR_AV_SP', CAR_AV * (SP != 0))
TRAIN_AV_SP = database.DefineVariable('TRAIN_AV_SP', TRAIN_AV * (SP != 0))

av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}


# Class membership model
W_OTHER = Beta('W_OTHER', 0.798, 0, 1, 0)
probClass1 = 1 - W_OTHER
probClass2 = W_OTHER

# The choice model is a discrete mixture of logit, with availability conditions
prob1 = PanelLikelihoodTrajectory(models.logit(V1, av, CHOICE))
prob2 = PanelLikelihoodTrajectory(models.logit(V2, av, CHOICE))
probIndiv = probClass1 * prob1 + probClass2 * prob2
logprob = log(MonteCarlo(probIndiv))


class test_15(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(database, logprob, numberOfDraws=5, seed=10)
        biogeme.saveIterations = False
        biogeme.generateHtml = False
        biogeme.generatePickle = False
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike, -4074.4822049451564, 2)


if __name__ == '__main__':
    unittest.main()



""" File: b07estimation_specification.py

 Author: Michel Bierlaire, EPFL
 Date: Thu Apr 13 21:04:47 2023

Estimation of a mixtures of logit models where the integral is
approximated using MonteCarlo integration.

"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, MonteCarlo, log

from swissmetro import (
    database,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    TRAIN_AV_SP,
    SM_AV,
    CAR_AV_SP,
    CHOICE,
)

R = 2000


def get_biogeme(the_draws):
    """Function returning the Biogeme object as a function of the selected draws

    :param the_draws: expression representing the draws
    :type the_draws: biogeme.expressions.bioDraws
    """

    ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
    ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
    B_TIME = Beta('B_TIME', 0, None, None, 0)
    B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
    B_COST = Beta('B_COST', 0, None, None, 0)

    # Define a random parameter, normally distributed, designed to be used
    # for Monte-Carlo simulation
    b_time_rnd = B_TIME + B_TIME_S * the_draws

    # Definition of the utility functions
    v_1 = ASC_TRAIN + b_time_rnd * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
    v_2 = b_time_rnd * SM_TT_SCALED + B_COST * SM_COST_SCALED
    v_3 = ASC_CAR + b_time_rnd * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

    # Associate utility functions with the numbering of alternatives
    V = {1: v_1, 2: v_2, 3: v_3}

    # Associate the availability conditions with the alternatives
    av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

    # The choice model is a logit, with availability conditions
    prob = models.logit(V, av, CHOICE)
    logprob = log(MonteCarlo(prob))

    the_biogeme = bio.BIOGEME(database, logprob)
    return the_biogeme



import fileinput
import os


def replace_in_files(file_list, search_string, replacement_string):
    for file_path in file_list:
        # Create a backup of the original file
        backup_file_path = file_path + '.bak'
        os.rename(file_path, backup_file_path)

        # Open the original file for reading and a temporary file for writing
        with fileinput.input(backup_file_path, inplace=True) as file, open(
            file_path, 'w'
        ) as temp_file:
            for line in file:
                # Replace the search string with the replacement string
                updated_line = line.replace(search_string, replacement_string)
                # Write the updated line to the temporary file
                temp_file.write(updated_line)

        # Remove the backup file
        os.remove(backup_file_path)


# Example usage
file_list = [
    'b01logit.py',
    'b02weight.py',
    'b03scale.py',
    'b05normal_mixture.py',
    'b05normal_mixture_integral.py',
    'b06unif_mixture.py',
    'b06unif_mixture_MHLS.py',
    'b06unif_mixture_integral.py',
    'b07discrete_mixture.py',
    'b08boxcox.py',
    'b09nested.py',
    'b10nested_bottom.py',
    'b11cnl.py',
    'b11cnl_sparse.py',
    'b12panel.py',
    'b12panel_flat.py',
    'b14nested_endogenous_sampling.py',
    'b15panel_discrete.py',
    'b15panel_discrete_bis.py',
    'b16panel_discrete_socio_eco.py',
    'b17lognormal_mixture.py',
    'b17lognormal_mixture_integral.py',
    'b18ordinal_logit.py',
    'b23logit.py',
    'b23probit.py',
    'b24halton_mixture.py',
    'b25triangular_mixture.py',
    'b26triangular_panel_mixture.py',
]

search_string = 'shortSummary'  # String to be replaced
replacement_string = 'short_summary'  # String to replace with

replace_in_files(file_list, search_string, replacement_string)



"""File b01logit_p.py

:author: Michel Bierlaire, EPFL
:date: Fri Apr 14 14:17:32 2023

 Example of a logit model.
The Swissmetro data is organized such that each row contains all the
responses of one individual.

"""
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, bioMultSum, Variable

from swissmetro import (
    database,
    NBR_QUESTIONS,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    TRAIN_AV_SP,
    SM_AV,
    CAR_AV_SP,
)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)


# Definition of the utility functions
V1 = [
    ASC_TRAIN + B_TIME * TRAIN_TT_SCALED[q] + B_COST * TRAIN_COST_SCALED[q]
    for q in range(NBR_QUESTIONS)
]
V2 = [
    ASC_SM + B_TIME * SM_TT_SCALED[q] + B_COST * SM_COST_SCALED[q]
    for q in range(NBR_QUESTIONS)
]
V3 = [
    ASC_CAR + B_TIME * CAR_TT_SCALED[q] + B_COST * CAR_CO_SCALED[q]
    for q in range(NBR_QUESTIONS)
]

# Associate utility functions with the numbering of alternatives
V = [{1: V1[q], 2: V2[q], 3: V3[q]} for q in range(NBR_QUESTIONS)]

# Associate the availability conditions with the alternatives
av = [{1: TRAIN_AV_SP[q], 2: SM_AV[q], 3: CAR_AV_SP[q]} for q in range(NBR_QUESTIONS)]

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = [
    models.loglogit(V[q], av[q], Variable(f'CHOICE_{q}')) for q in range(NBR_QUESTIONS)
]

# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, bioMultSum(logprob))
the_biogeme.modelName = '01logit_p'

# Estimate the parameters
results = the_biogeme.estimate()
print(results.shortSummary())
# Get the results in a pandas table
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File b08selected_specification.py

:author: Michel Bierlaire, EPFL
:date: Sat Jul 15 15:46:56 2023

We consider the model with 432 specifications:
- 3 models
    - logit
    - nested logit with two nests: public and private transportation
    - nested logit with two nests existing and future modes
- 3 functional form for the travel time variables
    - linear specification,
    - Box-Cox transform,
    - power series,
- 2 specification for the cost coefficients:
    - generic
    - alternative specific
- 2 specification for the travel time coefficients:
    - generic
    - alternative specific
- 4 segmentations for the constants:
    - not segmented
    - segmented by GA (yearly subscription to public transport)
    - segmented by luggage
    - segmented both by GA and luggage
-  3 segmentations for the time coefficients:
    - not segmented
    - segmented with first class
    - segmented with trip purpose

This leads to a total of 432 specifications.
After running the assisted specification algorithm, we select one
specification and estimate it.

"""
import biogeme.logging as blog
import biogeme.biogeme as bio
from everything_spec import model_catalog, database, av

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b07everything_assisted')

SPEC_ID = (
    'ASC:GA-LUGGAGE;'
    'B_COST_gen_altspec:generic;'
    'B_TIME:FIRST;'
    'B_TIME_gen_altspec:generic;'
    'model_catalog:logit;'
    'train_tt_catalog:power'
)

# The biogeme object can be obtained from the spec_id, and used as usual.
the_biogeme = bio.BIOGEME.from_configuration(
    config_id=SPEC_ID,
    expression=model_catalog,
    database=database,
)
the_biogeme.modelName = 'my_favorite_model'

the_biogeme.calculateNullLoglikelihood(av)

# Estimate the parameters
results = the_biogeme.estimate()
print(results.short_summary())

# Get the results in a pandas table
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File b02weight.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:02:59 2023

 Example of a logit model with Weighted Exogenous Sample Maximum
 Likelihood (WESML).
"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta

from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    GROUP,
)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# Definition of the weight
weight = 8.890991e-01 * (1.0 * (GROUP == 2) + 1.2 * (GROUP == 3))

# These notes will be included as such in the report file.
USER_NOTES = (
    'Example of a logit model with three alternatives: '
    'Train, Car and Swissmetro.'
    ' Weighted Exogenous Sample Maximum Likelihood estimator (WESML)'
)


# Create the Biogeme object
formulas = {'loglike': logprob, 'weight': weight}
the_biogeme = bio.BIOGEME(database, formulas, userNotes=USER_NOTES)
the_biogeme.modelName = 'b02weight'

# It is possible to control the generation of the HTML and the pickle
# files Note that these parameters can also be modified in the .TOML
# configuration file
the_biogeme.generate_html = True
the_biogeme.generate_pickle = False
results = the_biogeme.estimate()
print(results.short_summary())

# Get the results in a pandas table
pandas_results = results.getEstimatedParameters()
print(pandas_results)



"""File b22multiple_models.py

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 17:05:40 2023

 Example of the estimation of several versions of the model. In this
case, the number fo specifications exceed the maximum limit, so a heuristic is applied.
It may actually end up enumerating all possiblities.
"""

import biogeme.logging as blog
from biogeme.results import compile_estimation_results
from biogeme.multiobjectives import AIC_BIC_dimension
from biogeme.assisted import AssistedSpecification
from b22multiple_models_spec import the_biogeme, PARETO_FILE_NAME


logger = blog.get_screen_logger(blog.INFO)
logger.info('Example b22multiple_models')


nbr = the_biogeme.loglike.number_of_multiple_expressions()
if nbr is None:
    print('There are too many possible specifications to be enumerated')
else:
    print(f'There are {nbr} possible specifications')
assisted_specification = AssistedSpecification(
    biogeme_object=the_biogeme,
    multi_objectives=AIC_BIC_dimension,
    pareto_file_name=PARETO_FILE_NAME,
)

non_dominated_models = assisted_specification.run()

summary, description = compile_estimation_results(
    non_dominated_models, use_short_names=True
)
print(summary)
for k, v in description.items():
    if k != v:
        print(f'{k}: {v}')



"""File airline_results.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 16:26:35 2023

Assisted specification for the airline case study. Generate results
from the pareto file.

"""
try: 
    import matplotlib.pyplot as plt
    can_plot = True
except ModuleNotFoundError:
    can_plot = False

import biogeme.logging as blog
from biogeme.assisted import AssistedSpecification, ParetoPostProcessing
from biogeme.results import compile_estimation_results
from biogeme.multiobjectives import loglikelihood_dimension

from airline_spec import the_biogeme

screen_logger = blog.get_screen_logger(blog.INFO)
screen_logger.info('airline example: assisted specification')

PARETO_FILE_NAME = 'airline.pareto'

assisted_specification = AssistedSpecification(
    biogeme_object=the_biogeme,
    multi_objectives=loglikelihood_dimension,
    pareto_file_name=PARETO_FILE_NAME,
)

post_processing = ParetoPostProcessing(
    biogeme_object=the_biogeme,
    pareto_file_name=PARETO_FILE_NAME,
)
post_processing.log_statistics()

all_results = post_processing.reestimate(recycle=True)

summary, description = compile_estimation_results(all_results, use_short_names=True)
print(summary)
for k, v in description.items():
    if k != v:
        print(f'{k}: {v}')

if can_plot:
    ax = post_processing.plot()
    plt.show()



"""File b08boxcox.py

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:58:15 2023

 Example of a logit model, with a Box-Cox transform of variables.

"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)
LAMBDA = Beta('LAMBDA', 0, None, None, 0)

# Definition of the utility functions
V1 = (
    ASC_TRAIN
    + B_TIME * models.boxcox(TRAIN_TT_SCALED, LAMBDA)
    + B_COST * TRAIN_COST_SCALED
)
V2 = ASC_SM + B_TIME * models.boxcox(SM_TT_SCALED, LAMBDA) + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * models.boxcox(CAR_TT_SCALED, LAMBDA) + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)


# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b08boxcox'

the_biogeme.checkDerivatives(beta=[0, 0, 0, 0, 0], verbose=True)

# Estimate the parameters
results = the_biogeme.estimate()
print(results.short_summary())
pandas_results = results.getEstimatedParameters()
print(pandas_results)
